"""
Google Sheets provides a good platform to curate tabular data collaboratively. Thus, we
can use it to work on input for CLDF datasets, such as language metadata. This module provides
functionality to pull such data from Google Sheets, suitable for inclusion in `cldfbench download`
command implementations.

We use the `gspread <https://pypi.org/project/gspread/>`_ library to interact with data in
Google Sheets. Thus, code needs access to a JSON OAuth2 key file to authenticate it.
Retrieving such a key file (and giving permissions to the corresponding service account
for particular sheets) is described at `<https://docs.gspread.org/en/latest/oauth2.html>`_.

Thus, everyone running `cldfbench download` for a dataset using Google Sheets, would need
(access to) a JSON OAuth2 key file with appropriate authorizations.
"""
import typing
import pathlib
import collections

from csvw.dsv import UnicodeWriter
from collabutils.util import warn
try:
    import gspread
except ImportError as e:  # pragma: no cover
    warn('googlesheets.Spreadsheet', 'googlesheets', e)
from collabutils.base import SharedSpreadsheetMixin

__all__ = ['Spreadsheet']


def get_service_account(keyfile=None):
    """
    https://docs.gspread.org/en/latest/oauth2.html#authentication

    :param keyfile:
    :return:
    """
    keyfile = keyfile or input('Path to Google API key file: ')
    keyfile = pathlib.Path(keyfile)
    assert keyfile.exists()
    return gspread.service_account(
        filename=str(keyfile), scopes=['https://spreadsheets.google.com/feeds'])


class Spreadsheet(SharedSpreadsheetMixin):
    """
    Typical usage in a `cldfbench.Dataset`'s `cmd_download` method:

    .. code-block:: python

        >>> document = Spreadsheet('doc-key', auth='credentials.json')
        >>> document.fetch_sheets(
        ...     sheets={'Varieties': 'languages.tsv'})
        ...     outdir=dataset.etc_dir,
        ...     delimiter='\t')
    """

    def __init__(
            self,
            key_or_url: str,
            auth: typing.Optional[typing.Union[pathlib.Path, str]] = None):
        service = get_service_account(keyfile=auth)
        if key_or_url.startswith('https://') or key_or_url.startswith('http://'):
            self.workbook = service.open_by_url(key_or_url)
        else:
            self.workbook = service.open_by_key(key_or_url)

    @property
    def sheets(self) -> typing.Dict[str, gspread.models.Spreadsheet]:
        return collections.OrderedDict([(s.title, s) for s in self.workbook.worksheets()])

    def fetch_sheets(
            self,
            sheets: typing.Optional[typing.Dict[str, str]] = None,
            outdir: typing.Optional[typing.Union[pathlib.Path, str]] = '.',
            **kw,
    ):
        """
        :param sheets: A mapping of sheet titles to filenames, specifying which sheets to save \
        in which files. If None, all sheets will be saved to files "sheet_<no>.csv".
        :param outdir:
        :param kw: Passed as kwargs into UnicodeWriter.
        """
        for i, (title, sheet) in enumerate(self.sheets.items(), start=1):
            if (title in sheets) or sheets is None:
                fname = 'sheet_{}.csv'.format(i) if sheets is None else sheets[title]
                with UnicodeWriter(pathlib.Path(outdir) / fname, **kw) as writer:
                    writer.writerows(sheet.get_all_values())



"""
The `EDICTOR application <http://lingulist.de/edictor/>`_ is an online tool for creating,
maintaining, and publishing etymological data which is stored in simple TSV format.
This TSV format can be accessed in a somewhat configurable way via HTTP URLs.
"""
import typing
import pathlib
import urllib.parse
from urllib.request import urlretrieve


def get_url(
        base_url="https://lingulist.de/edictor",
        path=None,
        **params
):
    params = {
        k: '|'.join(v) if isinstance(v, (list, tuple)) else v
        for k, v in params.items() if v is not None}
    url_parts = list(urllib.parse.urlparse(base_url))
    if path:
        if not url_parts[2].endswith('/') and not path.startswith('/'):
            path = '/' + path
        url_parts[2] += path
    query = dict(urllib.parse.parse_qsl(url_parts[4]))
    query.update(params)
    url_parts[4] = urllib.parse.urlencode(query)
    return urllib.parse.urlunparse(url_parts)


def fetch(
        dataset: str,
        remote_dbase: typing.Optional[str] = None,
        concepts: typing.Optional[typing.List[str]] = None,
        languages: typing.Optional[typing.List[str]] = None,
        columns: typing.Optional[typing.List[str]] = None,
        base_url: str = "http://lingulist.de/edictor",
        out: typing.Optional[typing.Union[pathlib.Path, str]] = None,
):
    """
    Retrieve the TSV data file for a dataset curated with EDICTOR.

    :param dataset: Name of the dataset in EDICTOR
    :param remote_dbase: Name of the database file on EDICTOR (if different than dataset name)
    :param concepts: List of concepts to include in the download (defaults to all)
    :param languages: List of languages to include in the download (defaults to all)
    :param columns: List of columns to include in the download (defaults to all)
    :param base_url:
    :param out: Path to which to save the data (defaults to './<dataset>.tsv')
    :return:
    """
    remote_dbase = remote_dbase or (dataset + '.sqlite')
    url = get_url(
        base_url=base_url,
        path='/triples/get_data.py',
        file=dataset,
        concepts=concepts,
        doculects=languages,
        columns=columns,
        remote_dbase=remote_dbase,
    )
    if out:
        out = pathlib.Path(out)
        if out.is_dir():
            out = out / '{}.tsv'.format(dataset)
    else:
        out = pathlib.Path('{}.tsv'.format(dataset))
    return urlretrieve(url, str(out))



import warnings


def warn(name, extra, e):
    warnings.warn(
        '{0} is only available when collabutils is installed with {1} support\n'
        'pip install collabutils[{1}]'.format(name, extra))
    raise e



"""
OwnCloud compatible cloud services (e.g. NextCloud) often offer good support for collaborative
editing of tabular data in an online replica of Excel.

Public read-only shares provide a simple way to access such data programmatically.
"""
import typing
import pathlib
import tempfile
import collections

from csvw.dsv import UnicodeWriter
from collabutils.util import warn
try:
    import owncloud
    import openpyxl
except ImportError as e:  # pragma: no cover
    warn('oc.Spreadsheet', 'owncloud', e)

from collabutils.base import SharedSpreadsheetMixin

__all__ = ['Spreadsheet']

File = collections.namedtuple('File', 'name path content_type'.split())


class Share:
    def __init__(self, link):
        self.client = owncloud.Client.from_public_link(link)

    @property
    def files(self):
        return [File(f.name, f.path, f.get_content_type()) for f in self.client.list('')]


class Spreadsheet(SharedSpreadsheetMixin):
    """
    Typical usage in a `cldfbench.Dataset`'s `cmd_download` method:

    .. code-block:: python

        >>> document = Spreadsheet(
        ...     'test.xlsx', 'https://share.eva.mpg.de/index.php/s/pGaomxWqHPqxeEA')
        >>> document.fetch_sheets(
        ...     sheets={'Varieties': 'languages.tsv'})
        ...     outdir=dataset.etc_dir,
        ...     delimiter='\t')
    """
    def __init__(self, fname, share_link):
        share = Share(share_link)
        for f in share.files:
            if f.name == fname:
                self.file = f
                break
        else:
            raise ValueError(fname)
        self.client = share.client

    def fetch_sheets(
            self,
            sheets: typing.Optional[typing.Dict[str, str]] = None,
            outdir: typing.Optional[typing.Union[pathlib.Path, str]] = '.',
            **kw,
    ):
        import shutil

        def _excel_value(x):
            if x is None:
                return ""
            if isinstance(x, float) and int(x) == x:
                # Since Excel does not have an integer type, integers are rendered as "n.0",
                # which in turn confuses type detection of tools like csvkit. Thus, we normalize
                # numbers of the form "n.0" to "n".
                return '{0}'.format(int(x))  # pragma: no cover
            return '{0}'.format(x).strip()

        with tempfile.TemporaryDirectory() as tmp:
            tmppath = str(pathlib.Path(tmp) / self.file.name)
            self.client.get_file(self.file.path, tmppath)
            shutil.copy(tmppath, 'lexibank.xlsx')
            outdir = outdir or self
            wb = openpyxl.load_workbook(tmppath, data_only=True)
            print(wb.sheetnames)
            for i, sname in enumerate(wb.sheetnames, start=1):
                if sheets is None or (sname in sheets):
                    fname = 'sheet_{}.csv'.format(i) if sheets is None else sheets[sname]
                    sheet = wb[sname]
                    with UnicodeWriter(pathlib.Path(outdir) / fname, **kw) as writer:
                        for row in sheet.rows:
                            writer.writerow([_excel_value(col.value) for col in row])



import typing
import pathlib


class SharedSpreadsheetMixin:
    def fetch_sheets(
            self,
            sheets: typing.Optional[typing.Dict[str, str]] = None,
            outdir: typing.Optional[typing.Union[pathlib.Path, str]] = '.',
            **kw,
    ):
        raise NotImplementedError()  # pragma: no cover



"""
Zotero provides a good platform to collaborate on bibliogrpahies.

We support collaborative curation of bibliographical data as follows:

- The data is curated in a shared Zotero library.
- Items in this library are tagged for membership in the bibliographies of certain datasets, \
  using tags of the form "dataset:<dataset-id>".
- Items are optionally tagged with tags of the form "id:<dataset-id>:<local-id>" for local \
  BibTeX keys in a specific dataset.

An instance of :class:`Zotero` allows

- seeding such libraries from a (possibly stub) BibTeX file via :meth:`Zotero.upload_bib`
- downloading such libraries via :meth:`Zotero.download_bib`
- deleting such libraries via :meth:`Zotero.delete_bib`
"""
import re
import copy
import pathlib
import collections

from pycldf.sources import Source

from collabutils.util import warn
try:
    from pybtex.database import parse_string
    from pyzotero.zotero import Zotero as API
except ImportError as e:  # pragma: no cover
    warn('zotero.Zotero', 'zotero', e)

__all__ = ['Zotero']

BIBTEX_TO_ZOTERO = {
    'article': 'journalArticle',
    'book': 'book',
    'booklet': 'book',
    'conference': 'conferencePaper',
    'inbook': 'bookSection',
    'incollection': 'bookSection',
    'inproceeding': 'conferencePaper',
    # manual
    'mastersthesis': 'thesis',
    'misc': 'document',
    'phdthesis': 'thesis',  # + thesisType! 'PhD thesis' or 'masters thesis'
    # proceeding
    'techreport': 'report',
    'unpublished': 'manuscript',
    # BibLaTeX
    # mvbook
    # bookinbook
    # suppbook
    # booklet
    # collection
    # mvcollection
    # incollection
    # suppcollection
    # online
    # patent
    # periodical
    # suppperiodical
    # mvproceedings
    # reference
    # mvreference
    # inreference
    # report
    # set
    'thesis': 'thesis',
    # custom
    # electronic
}

FIELDS_TO_ZOTERO = {
    'address': 'place',
    # 'author -> creators
    'booktitle': 'bookTitle',
    'doi': 'DOI',
    # 'editor -> creators
    'journal': 'publicationTitle',
    'school': 'institution',
    'year': 'date',
    'abstract': 'abstractNote',
}


class Zotero:
    """
    A high-level Zotero API client. Low level communication with Zotero is done using the
    `pyzotero` package.

    .. seealso:: `<https://pypi.org/project/Pyzotero/>`_
    """
    def __init__(self, libid, apikey, group=True):
        """
        :param libid: Numeric ID of the Zotero library
        :param apikey: API key with read/write permissions for the library
        :param group: Flag signaling whether we are dealing with a group or user library
        """
        self.api = API(libid, 'group' if group else 'user', apikey)
        self._item_templates = {}

    @staticmethod
    def id_tag(dataset_id, bibkey):
        return 'id:{}:{}'.format(dataset_id, bibkey)

    @staticmethod
    def id_from_tagstring(s, dataset_id):
        m = re.search(r'id:{}:(?P<id>[^\s,]+)'.format(re.escape(dataset_id)), s)
        if m:
            return m.group('id')

    @staticmethod
    def dataset_tag(dataset_id):
        return 'dataset:{}'.format(dataset_id)

    def upload_bib(self, dataset_id, bibpath, log=None):
        """
        Upload items from a BibTeX file as items of the Zotero library for a dataset.

        :param dataset_id: Dataset identifier (e.g. a `cldfbench.Dataset.id`).
        :param bibpath: Path to the BibTeX file to be uploaded.
        :param log:
        :return:
        """
        res = collections.OrderedDict()
        db = parse_string(pathlib.Path(bibpath).read_text(encoding='utf8'), 'bibtex')
        for key, entry in db.entries.items():
            src = Source.from_entry(key, entry)
            tag = self.id_tag(dataset_id, src.id)
            try:
                _ = self.api.items(tag=tag, limit=1)[0]
                if log:  # pragma: no cover
                    log.info('Skipping {}'.format(tag))
                res[src.id] = True
            except IndexError:
                res[src.id] = self.create_item(src, tag, self.dataset_tag(dataset_id))
        return res

    def download_bib(self, dataset_id, bibpath=None, remove=None):
        """
        Download all items tagged for a dataset into a BibTeX file.

        :param dataset_id:
        :param bibpath:
        :param remove: List of field names to remove (e.g. the "keywords" field, which is popuated \
        by Zotero's BibTeX export with all tags) or `None`.
        :return:
        """
        remove = remove or []
        bib = []
        for item in self.get_items(self.dataset_tag(dataset_id), content='bibtex'):
            src = self._bibtex2source(item)
            # determine the dataset-local bibkey:
            id_ = self.id_from_tagstring(src.get('keywords', ''), dataset_id)
            if id_:
                src.id = id_
            for k in remove:
                if k in src:
                    del src[k]
            bib.append(src.bibtex())
        bib = '\n\n'.join(bib)
        if bibpath:
            pass
        return bib

    def delete_bib(self, dataset_id, log=None):
        """
        Delete all items tagged for a dataset - unless they are tagged for other datasets as well.

        :param dataset_id:
        :param log:
        :return:
        """
        for item in self.get_items(self.dataset_tag(dataset_id)):
            if any(':' + dataset_id not in t['tag'] for t in item['data']['tags']):
                if log:
                    log.warning('Skipping item linked to other datasets')
                continue
            self.api.delete_item(item)

    def get_items(self, tag, **kw):
        return self.api.everything(self.api.items(tag=tag, **kw))

    @staticmethod
    def _bibtex2source(bibtex):
        src = Source.from_entry(*list(parse_string(bibtex, 'bibtex').entries.items())[0])

        def repl(m):
            if m.group('cat') == 'type':
                src.genre = m.group('value').strip()
            else:
                field, value = m.group('value').split(':', maxsplit=1)
                if field == 'note':
                    return value.strip()
                src[field] = value.strip()
            return ''

        if 'note' in src:
            note = re.sub(
                ':bibtex:(?P<cat>type|field):(?P<value>.+?):xetbib:', repl, src['note']).strip()
            if note:
                src['note'] = note
            else:
                del src['note']
        return src

    def _empty_item(self, src):
        zot_genre = BIBTEX_TO_ZOTERO.get(src.genre.lower(), 'document')
        if zot_genre not in self._item_templates:
            self._item_templates[zot_genre] = self.api.item_template(zot_genre)
        return copy.deepcopy(self._item_templates[zot_genre])

    def create_item(self, src, *tags):
        """
        """
        item = self._empty_item(src)

        extra = {'type': src.genre}
        for k, v in src.items():
            kl = k.lower()
            if kl in ['editor', 'author']:
                continue
            zotk = FIELDS_TO_ZOTERO.get(kl, kl)
            if zotk in item and isinstance(item[zotk], str):
                item[zotk] = src[k]
            else:
                extra['field:{}'.format(kl)] = src[k]

        if extra:
            item['extra'] = '\n'.join(
                [':bibtex:{}: {}:xetbib:'.format(k, v) for k, v in extra.items()])

        entry = src.entry
        item['creators'] = []
        for ct in ['author', 'editor']:
            for name in entry.persons[ct]:
                item['creators'].append(dict(
                    creatorType=ct,
                    firstName=str(name).split(',')[-1].strip(),
                    lastName=str(name).split(',')[0]))
        item['tags'].extend(tags)
        item['tags'] = [dict(tag=t) for t in item['tags']]
        return bool(self.api.create_items([item])['success'])
